"""The Python Session Control Library is used to control iTest sessions to enable quick and easy interaction with
devices from a Python script.

Target devices/APIs may be defined either on the fly, or via session profiles defined in iTest projects.

The library may operate in either one of these modes:

    * Connected to an iTest GUI or Velocity agent instance.
    * Standalone on a workstation. In this mode, the library runs its own instance of the agent, which should be
    installed locally.
"""



# ****************       BEGIN-STANDARD-COPYRIGHT      ***************
#
# Copyright (c) 2017, Spirent Communications.
#
# All rights reserved. Proprietary and confidential information of Spirent Communications.
#
#  ***************        END-STANDARD-COPYRIGHT       ***************

class UriIdentity(object):
    """ Base class for any resources in SLC with URI identity """
    def __init__(self, uri):
        """Initialize a URI identity object"""
        self._uri = uri

    @property
    def uri(self):
        """ Return a topology URI"""
        return self._uri



# ****************       BEGIN-STANDARD-COPYRIGHT      ***************
#
# Copyright (c) 2017, Spirent Communications.
#
# All rights reserved. Proprietary and confidential information of Spirent Communications.
#
#  ***************        END-STANDARD-COPYRIGHT       ***************

"""Project class and necessary tools."""

import re
import six
from .session_profile import SessionProfile
from .topology import Topology, DeviceData
from .resources import ParameterFile, ResponseMapFile
from .identity import UriIdentity

from .python_name import python_name

if six.PY2:
    from urllib import quote
else:
    from urllib.parse import quote

class Project(UriIdentity):
    """This class represents a single project.

    Objects of this class should not be created directly, but rather obtained from an SLC object.
    """

    def __init__(self, project_name, protocol_socket, agent_type):
        UriIdentity.__init__(self, 'project://%s/' % str(project_name))
        self.project_name = project_name
        self._protosock = protocol_socket
        self.agent_type = agent_type

        self._update_agent_type()

        self._session_profiles = dict()
        self._topologies = dict()
        self._parameter_files = dict()
        self._response_maps = dict()

    def __str__(self):
        """Return a project name"""
        return self.project_name

    def open(self):
        """
        Parses itar and init a list of all the usable topologies and session profiles in the project.

        Returns self for convenience.
        """

        self._update_project_list()

        return self

    def _resource_name(self, itest_name, suffix):
        pos_from = itest_name.rfind('/') + 1
        pos_to = itest_name.rfind('.')
        return self._python_name(itest_name[pos_from:pos_to] + suffix)

    def _python_name(self, name):
       return python_name(name)

    def _update_agent_type(self):
        """ Detect agent type based on init packet send by server """
        pass

    def _update_project_list(self):
        ret = self._protosock.query_project(self.project_name, parameter_file=True, response_map=True)

        def _as_list(value, new_value):
            if value == None:
                value = new_value
            elif type(value) == list:
                value.append(new_value)
            else:
                value = [value, new_value]
            return value

        def _find(value, uri):
            if value != None:
                if isinstance(value, list):
                    for resource in value:
                        if resource.uri == uri:
                            return True
                else:
                    # value instanceof SessionProfile, Topology, ParameterFile or ResponseMapFile
                    if value.uri == uri:
                        return True

            return False

        def _check_prefix(value):
            if value.startswith('/'):
                return value[1:]
            return value

        for session_profile in ret.sessionProfiles:
            name = self._resource_name(session_profile, '_ffsp')
            uri = 'project://%s/%s' %(quote(self.project_name), quote(_check_prefix(session_profile)))

            value = self._session_profiles.get(name)

            # Check existance
            if _find(value, uri):
                continue
            # Update new value
            new_value = SessionProfile(self._protosock, self.agent_type, uri)
            self._session_profiles[name] = _as_list(value, new_value)

        for topology in ret.topologies:
            name = self._resource_name(topology.name, '_tbml')
            topology_uri = 'project://%s/%s' %(quote(self.project_name), quote(_check_prefix(topology.name)))

            value = self._topologies.get(name)
            if _find(value, topology_uri):
                continue

            devices = [DeviceData(self._python_name(device.name), quote(device.name), device.sessions) for device in topology.devices]
            new_value = Topology(self, topology_uri, devices, self._protosock)

            self._topologies[name] = _as_list(value, new_value)

        for param_file in ret.parameterFiles:
            name = self._resource_name(param_file, '_ffpt')
            uri = 'project://%s/%s' %(quote(self.project_name), quote(_check_prefix(param_file)))

            value = self._parameter_files.get(name)
            if _find(value, uri):
                continue
            new_value = ParameterFile(uri)
            self._parameter_files[name] = _as_list(value, new_value)

        for response_file in ret.responseMaps:
            name = self._resource_name(response_file, '_ffrm')
            uri = 'project://%s/%s' %(quote(self.project_name), quote(_check_prefix(response_file)))

            value = self._response_maps.get(name)
            if _find(value, uri):
                continue

            new_value = ResponseMapFile(uri)
            self._response_maps[name] = _as_list(value, new_value)

    def list(self, session_file=True, parameter_file=False, response_map=False, topology_file=True):
        """
        Returns a list of names of all the usable topologies and session profiles in the project.

        Resources with duplicate names will be shown once,
        but all access methods will return a list of resources with same name.

        Filter parameters:
        session_file -- include session files
        parameter_file -- include parameter files
        response_map -- include response map files
        topology_file -- include topology files

        Example:
            proj.list()
            ==> ['dut1_ffsp', 'lab1_setup_tbml']
        """
        self._update_project_list()

        result = list()
        if session_file:
            result += list(self._session_profiles.keys())

        if topology_file:
            result += list(self._topologies.keys())

        if parameter_file:
            result += list(self._parameter_files.keys())

        if response_map:
            result += list(self._response_maps.keys())

        return result

    def __dir__(self):
        """ return a list of methods available to project"""
        res = super.__dir__(self) + self.list(session_file=True, parameter_file=True, response_map=True, topology_file=True)
        return res

    def response_maps(self):
        """
        Returns a list of reseponse maps.

        Example:
            proj.response_maps()
            ==> ['dut1_ffrm', 'lab1_setup_ffrm']
        """
        return self.list(response_map=True, session_file=False, topology_file=False)

    def parameter_files(self):
        """
        Returns a list of parameter files

        Example:
            proj.response_maps()
            ==> ['dut1_ffrm', 'lab1_setup_ffrm']
        """
        return self.list(parameter_file=True, session_file=False, topology_file=False)

    def __getitem__(self, key):
        """
            Sessions are opened either directly on a session profile or local topology.

            Examples:
                s1 = proj.dut1_ffsp.open()
                s1 = proj.rest_session_ffsp.open(url='https://my_site.my_domain.com', accept_all_cookies=True)
                s1 = proj.rest_session_ffsp.open(parameter_file=proj.main_setup_ffpt)
                s1 = proj.rest_session_ffsp.open(
                    properties={'authentication.authentication': 'Basic', 'authentication.user': 'me', 'authentication.password': 'totes_secret!'})

            Opening native sessions:
                s1 = slc.sessions.ssh.open(ip_address='10.20.30.40')
        """
        profile = self._session_profiles.get(key)
        if profile:
            return profile

        topology = self._topologies.get(key)
        if topology:
            return topology

        parameter = self._parameter_files.get(key)
        if parameter:
            return parameter

        response_map = self._response_maps.get(key)
        if response_map:
            return response_map

        return None

    def __getattr__(self, key):
        ret = self[key]
        if ret is None:
            raise AttributeError(key)
        return ret



# ****************       BEGIN-STANDARD-COPYRIGHT      ***************
#
# Copyright (c) 2017, Spirent Communications.
#
# All rights reserved. Proprietary and confidential information of Spirent Communications.
#
#  ***************        END-STANDARD-COPYRIGHT       ***************

from SpirentSLC import SLC
import sys

with SLC.init(itar_path='itars/') as slc:
    print(slc.list())

    proj = slc.open('my_project')
    print(proj.list(parameter_file=True, response_map=True))

    # -------------------- Python session ---------------------------
    print('checking python version in iTest,\n current version is:' + sys.version)
    with proj.slc_test_ffsp.open() as s1:
        s1.command('import sys')
        print('python version in Agent is:' + s1.command('sys.version').text)

        print(s1.my_quickcall(param1=123, param2=True))

    #print("check telnet")
    # -------------------- Telnet session (somewhat broken atm) -----
    # with proj.telnet_ffsp.open(ipAddress='40.76.76.181', port=80) as s1:
    #     print('==== agent ====')
    #     print(s1.agent)

    # print("check HTTP")
    # # -------------------- HTTP session -----------------------------
    # with proj.http_ffsp.open() as s1:
    #     resp = s1.command('get page http://xored.com')
    #     print(resp.text)

    # -------------------- REST session -----------------------------
    # with proj.REST_ffsp.open(url='https://jsonplaceholder.typicode.com/') as s1:
    #     print(s1.GET('/posts/1'))




from SpirentSLC import SLC

with SLC.init(host='localhost:9005') as slc:
    print('============== Prolects list ==============')
    print(slc.list())

    print('============== NativePyLib project resources ==============')
    proj = slc.open('NativePyLib')
    print(proj.list(parameter_file=True, response_map=True))

    print('============== Quick calls ==============')
    print(proj.dut_ffsp.list())
    print('===> access the list of parameters for a specific QuickCall')
    print(proj.dut_ffsp.list('init_routes'))

    print('============== Topology ==============')
    with proj.ServerAndPC_tbml.server1.cmd.open() as cmd_session:
        print(cmd_session.command('help'))

    with proj.ServerAndPC_tbml.server1.Process.open() as process_session:
        print(process_session.list())




# ****************       BEGIN-STANDARD-COPYRIGHT      ***************
#
# Copyright (c) 2017, Spirent Communications.
#
# All rights reserved. Proprietary and confidential information of Spirent Communications.
#
#  ***************        END-STANDARD-COPYRIGHT       ***************

from xml.etree import ElementTree
from SpirentSLC import SLC

##
with SLC.init(host='localhost:9005', itar_path='itars') as slc:

    ## Openin a response map projecy
    proj = slc.open('response_map_test_prj')

    proj2 = slc.open('my_project')
    with proj2.slc_test_ffsp.open() as s2:
        # for i in range(1, 1000):
        # resp = s2.command('print(1)')
        resp = s2.eval('puts(1)')
        print(resp.text)

    # -------------------- HTTP session -----------------------------
    with proj.http_ffsp.open() as s1:

        for i in range(1, 5):
            resp = s1.command('get page http://www.w3schools.com/xml/cd_catalog.xml', response_map="xml_catalog.ffrm")
            print(resp.queries())

            ## Query for titles
            if resp.xml != None:
                print([x.text for x in resp.xml.findall('.//TITLE')])

            print('do same with custom query')
            print(resp.titles())

    # -------------------- REST session -----------------------------
    print('rest session')
    with slc.sessions.rest.open(url='http://jsonplaceholder.typicode.com/') as s1:
        print(s1.GET('/posts/1').text)
    print('Done')


# ****************       BEGIN-STANDARD-COPYRIGHT      ***************
#
# Copyright (c) 2017, Spirent Communications.
#
# All rights reserved. Proprietary and confidential information of Spirent Communications.
#
#  ***************        END-STANDARD-COPYRIGHT       ***************

"""Session profile class and necessary tools."""

import re
import six
import json

from xml.etree import ElementTree

if six.PY2:
    import codecs

from .internal import ExecutionProtocol_pb2 as pb


_DEBUG = False

_RESULT_MESSAGE = {
    pb.SUCCESS: 'success',
    pb.ABORTED: 'aborted',
    pb.TIMEOUT: 'timeout',
    pb.FAILED: 'failed',
    pb.SKIPPED: 'skipped'
}

_FIRST_CAP_RE = re.compile('(.)([A-Z][a-z]+)')
_ALL_CAP_RE = re.compile('([a-z0-9])([A-Z])')
_CAP2_PLUS = re.compile('(_)(_+)')

def _to_snake_name(name):
    temp_var = _FIRST_CAP_RE.sub(r'\1_\2', name)
    res = _ALL_CAP_RE.sub(r'\1_\2', temp_var).lower()
    res = res.replace('-', '_')
    res = _CAP2_PLUS.sub(r'\1', res)
    return res

class SessionActionResponse(object):
    """ A wrapper around a session action response """
    def __init__(self, message):
        self.message = message

        # Create dummy structured data
        self._xml = ElementTree.Element('structure')
        self._json_data = None
        self._xml_data = None
        # Need to convert all structured data to structure xml document
        self._prepare_structured_data()

        self._queries = {}
        try:
            if self.message.details != None:
                self._queries = { _to_snake_name(x.name): x for x in self.message.details.aliases }
        except AttributeError:
            pass


    def _prepare_structured_data(self):
        if not hasattr(self.message, 'details') or self.message.details is None:
            return

        mapped = None
        for s in self.message.details.structuredData:
            if '*' == s.map:
                for item in s.items:
                   self._item_to_element(self._xml, item)

            else:
                if mapped is None:
                    mapped = ElementTree.SubElement(self._xml, 'mapped')

                for item in s.items:
                    if item.name == 'json':
                        self._json_data = json.loads(self.message.details.body)
                    elif item.name == 'xml':
                        self._xml_data = ElementTree.fromstring(self._text().encode('utf-8'))

                    # every value item here is stored XML, so we need to parse it from this items
                    mapped_child = ElementTree.fromstring(item.value.encode('utf-8'))
                    mapped.append(mapped_child)

    def _item_to_element(self, root, item):
        item_element = ElementTree.SubElement(root, item.name)
        if len(item.items) > 0:
            for child_item in item.items:
                self._item_to_element(item_element, child_item)
        else:
            item_element.text = item.value
    def _result(self):
        """
            Return action response result
            May be 'success', 'failed', 'timeout', 'aborted', 'skipped'
        """
        res = _RESULT_MESSAGE[self.message.stepResult]
        if res is None:
            return 'unknown'
        return res

    def _text(self):
        """ return a text output of executed action """
        ret = None
        if hasattr(self.message, 'details') and self.message.details != None:
            if six.PY2:
                ret = codecs.encode(self.message.details.body, 'utf-8')
            else:
                ret = self.message.details.body
        if ret is None:
            return ''
        return ret

    def __str__(self):
        return self._text()

    def __repr__(self):
        return self._text()

    def _structure(self):
        """ Return structured data of response"""
        return self.message.details.structuredData

    @property
    def text(self):
        """ return a text output of executed action """
        return self._text()

    @property
    def json(self):
        """ return a json data"""
        return self._json_data

    @property
    def xml(self):
        """ return a xml data"""
        return self._xml_data

    def __getattr__(self, key):
        if key == 'text':
            return self._text()
        elif key == 'result':
            return self._result()
        elif key == 'structure':
            return self._structure()
        elif key == 'json':
            return self._json_data
        elif key == 'xml':
            return self._xml_data
        elif key == 'data':
            return self.message.details.structuredData
        else:
            ret = self[key]
            if ret != None:
                return ret
        raise AttributeError(key)

    def __getitem__(self, key):
        if key in self._queries:
            return lambda *args: self._invoke_query(key, *args)
        raise AttributeError(key)

    def _invoke_query(self, query_name, parameters=None):
        alias = self._queries[query_name]
        if alias is None:
            return None

        if _DEBUG:
            print("do query:", alias, " params: ", parameters)
        if alias.queryFormat != None:
            query = alias.queryFormat
            if not query.startswith('.//'):
                query = './/'+ query
            # Process arguments
            if parameters != None and len(alias.arguments) > 0:
                i = 0
                param_list = []
                if isinstance(parameters, list):
                    param_list = parameters

                param_list = [parameters,]
                for param in param_list:
                    query = query.replace('{'+str(i)+'}', str(param))
                    i += 1
            if _DEBUG:
                print("final query:", query)
            elements = self._xml.findall(query)
            if len(elements) == 1:
                return elements[0].text
            elif len(elements) > 1:
                if isinstance(parameters, int):
                    return elements[int(parameters)].text
                else:
                    return [x.text for x in elements]

        return None

    def queries(self):
        """
            Return a method names that query the structured data and return values.
            Queries may be auto-generated in iTest or be defined in response maps.
        """
        res = '['
        first_time = True
        keys = sorted(self._queries.keys())
        for x in keys:
            if first_time:
                first_time = False
            else:
                res += ', '
            msg = self._queries[x]
            if len(msg.arguments) > 0:
                # print(msg)
                res += '\'' + x + '('
                first_time_child = True
                for arg in msg.arguments:
                    if first_time_child:
                        first_time_child = False
                    else:
                        res +=', '
                    res += arg.name
                res +=  ')\''
            else:
                res += '\'' + x + '\''
        res += ']'
        return res



# ****************       BEGIN-STANDARD-COPYRIGHT      ***************
#
# Copyright (c) 2017, Spirent Communications.
#
# All rights reserved. Proprietary and confidential information of Spirent Communications.
#
#  ***************        END-STANDARD-COPYRIGHT       ***************

"""Python attribute name conversion utils"""
import re
def python_name(name):
    """Converts the file name to valid Python identifier - replaces the characters
    that are not legal in a Python identifier by underscores.
        identifier ::=  (letter|"_") (letter | digit | "_")*
        letter     ::=  lowercase | uppercase
        lowercase  ::=  "a"..."z"
        uppercase  ::=  "A"..."Z"
        digit      ::=  "0"..."9"
    """
    # replace invalid characters by underscores
    py_name = re.sub('[^0-9a-zA-Z_]', '_', name)
    # remove invalid leading (non-alphabet) characters
    return re.sub('^[^A-Za-z]*', '', py_name)
